# pylint:disable=ungrouped-imports

from unittest.mock import patch

import pytest

import auditor

from events.registry import tensorboard as tensorboard_events
from factories.factory_plugins import TensorboardJobFactory
from factories.factory_projects import ProjectFactory
from tests.test_auditor.utils import AuditorBaseTest


@pytest.mark.auditor_mark
class AuditorTensorboardTest(AuditorBaseTest):
    """Testing subscribed events"""
    EVENTS = tensorboard_events.EVENTS

    def setUp(self):
        self.tensorboard = TensorboardJobFactory(project=ProjectFactory())
        super().setUp()
        self.tested_events = {
            tensorboard_events.TENSORBOARD_STARTED,
            tensorboard_events.TENSORBOARD_STARTED_TRIGGERED,
            tensorboard_events.TENSORBOARD_STOPPED,
            tensorboard_events.TENSORBOARD_STOPPED_TRIGGERED,
            tensorboard_events.TENSORBOARD_CLEANED_TRIGGERED,
            tensorboard_events.TENSORBOARD_VIEWED,
            tensorboard_events.TENSORBOARD_UNBOOKMARKED,
            tensorboard_events.TENSORBOARD_BOOKMARKED,
            tensorboard_events.TENSORBOARD_NEW_STATUS,
            tensorboard_events.TENSORBOARD_FAILED,
            tensorboard_events.TENSORBOARD_SUCCEEDED,
            tensorboard_events.TENSORBOARD_STATUSES_VIEWED,
            tensorboard_events.TENSORBOARD_UPDATED,
            tensorboard_events.TENSORBOARD_DELETED,
            tensorboard_events.TENSORBOARD_DELETED_TRIGGERED,
            tensorboard_events.TENSORBOARD_ARCHIVED,
            tensorboard_events.TENSORBOARD_RESTORED,
        }

    @patch('executor.executor_service.ExecutorService.record_event')
    @patch('notifier.service.NotifierService.record_event')
    @patch('tracker.service.TrackerService.record_event')
    @patch('activitylogs.service.ActivityLogService.record_event')
    def test_tensorboard_started(self,
                                 activitylogs_record,
                                 tracker_record,
                                 notifier_record,
                                 executor_record):
        auditor.record(event_type=tensorboard_events.TENSORBOARD_STARTED,
                       instance=self.tensorboard,
                       target='project')

        assert tracker_record.call_count == 1
        assert activitylogs_record.call_count == 0
        assert notifier_record.call_count == 1
        assert executor_record.call_count == 1

    @patch('executor.executor_service.ExecutorService.record_event')
    @patch('notifier.service.NotifierService.record_event')
    @patch('tracker.service.TrackerService.record_event')
    @patch('activitylogs.service.ActivityLogService.record_event')
    def test_tensorboard_started_triggered(self,
                                           activitylogs_record,
                                           tracker_record,
                                           notifier_record,
                                           executor_record):
        auditor.record(event_type=tensorboard_events.TENSORBOARD_STARTED_TRIGGERED,
                       instance=self.tensorboard,
                       actor_id=1,
                       actor_name='foo',
                       target='project')

        assert tracker_record.call_count == 1
        assert activitylogs_record.call_count == 1
        assert notifier_record.call_count == 0
        assert executor_record.call_count == 0

    @patch('executor.executor_service.ExecutorService.record_event')
    @patch('notifier.service.NotifierService.record_event')
    @patch('tracker.service.TrackerService.record_event')
    @patch('activitylogs.service.ActivityLogService.record_event')
    def test_tensorboard_stopped(self,
                                 activitylogs_record,
                                 tracker_record,
                                 notifier_record,
                                 executor_record):
        auditor.record(event_type=tensorboard_events.TENSORBOARD_STOPPED,
                       instance=self.tensorboard,
                       target='project')

        assert tracker_record.call_count == 1
        assert activitylogs_record.call_count == 0
        assert notifier_record.call_count == 1
        assert executor_record.call_count == 1

    @patch('executor.executor_service.ExecutorService.record_event')
    @patch('notifier.service.NotifierService.record_event')
    @patch('tracker.service.TrackerService.record_event')
    @patch('activitylogs.service.ActivityLogService.record_event')
    def test_tensorboard_stopped_triggered(self,
                                           activitylogs_record,
                                           tracker_record,
                                           notifier_record,
                                           executor_record):
        auditor.record(event_type=tensorboard_events.TENSORBOARD_STOPPED_TRIGGERED,
                       instance=self.tensorboard,
                       actor_id=1,
                       actor_name='foo',
                       target='project')

        assert tracker_record.call_count == 1
        assert activitylogs_record.call_count == 1
        assert notifier_record.call_count == 0
        assert executor_record.call_count == 0

    @patch('executor.executor_service.ExecutorService.record_event')
    @patch('notifier.service.NotifierService.record_event')
    @patch('tracker.service.TrackerService.record_event')
    @patch('activitylogs.service.ActivityLogService.record_event')
    def test_tensorboard_cleaned_triggered(self,
                                           activitylogs_record,
                                           tracker_record,
                                           notifier_record,
                                           executor_record):
        auditor.record(event_type=tensorboard_events.TENSORBOARD_CLEANED_TRIGGERED,
                       instance=self.tensorboard,
                       actor_id=1,
                       actor_name='foo',
                       target='project')

        assert tracker_record.call_count == 0
        assert activitylogs_record.call_count == 0
        assert notifier_record.call_count == 0
        assert executor_record.call_count == 1

    @patch('executor.executor_service.ExecutorService.record_event')
    @patch('notifier.service.NotifierService.record_event')
    @patch('tracker.service.TrackerService.record_event')
    @patch('activitylogs.service.ActivityLogService.record_event')
    def test_tensorboard_viewed(self,
                                activitylogs_record,
                                tracker_record,
                                notifier_record,
                                executor_record):
        auditor.record(event_type=tensorboard_events.TENSORBOARD_VIEWED,
                       instance=self.tensorboard,
                       actor_id=1,
                       actor_name='foo',
                       target='project')

        assert tracker_record.call_count == 1
        assert activitylogs_record.call_count == 1
        assert notifier_record.call_count == 0
        assert executor_record.call_count == 0

    @patch('executor.executor_service.ExecutorService.record_event')
    @patch('notifier.service.NotifierService.record_event')
    @patch('tracker.service.TrackerService.record_event')
    @patch('activitylogs.service.ActivityLogService.record_event')
    def test_tensorboard_unbookmarked(self,
                                      activitylogs_record,
                                      tracker_record,
                                      notifier_record,
                                      executor_record):
        auditor.record(event_type=tensorboard_events.TENSORBOARD_UNBOOKMARKED,
                       instance=self.tensorboard,
                       actor_id=1,
                       actor_name='foo',
                       target='project')

        assert tracker_record.call_count == 1
        assert activitylogs_record.call_count == 1
        assert notifier_record.call_count == 0
        assert executor_record.call_count == 0

    @patch('executor.executor_service.ExecutorService.record_event')
    @patch('notifier.service.NotifierService.record_event')
    @patch('tracker.service.TrackerService.record_event')
    @patch('activitylogs.service.ActivityLogService.record_event')
    def test_tensorboard_bookmarked(self,
                                    activitylogs_record,
                                    tracker_record,
                                    notifier_record,
                                    executor_record):
        auditor.record(event_type=tensorboard_events.TENSORBOARD_BOOKMARKED,
                       instance=self.tensorboard,
                       actor_id=1,
                       actor_name='foo',
                       target='project')

        assert tracker_record.call_count == 1
        assert activitylogs_record.call_count == 1
        assert notifier_record.call_count == 0
        assert executor_record.call_count == 0

    @patch('executor.executor_service.ExecutorService.record_event')
    @patch('notifier.service.NotifierService.record_event')
    @patch('tracker.service.TrackerService.record_event')
    @patch('activitylogs.service.ActivityLogService.record_event')
    def test_tensorboard_new_status(self,
                                    activitylogs_record,
                                    tracker_record,
                                    notifier_record,
                                    executor_record):
        auditor.record(event_type=tensorboard_events.TENSORBOARD_NEW_STATUS,
                       instance=self.tensorboard,
                       target='project')

        assert tracker_record.call_count == 1
        assert activitylogs_record.call_count == 0
        assert notifier_record.call_count == 0
        assert executor_record.call_count == 1

    @patch('executor.executor_service.ExecutorService.record_event')
    @patch('notifier.service.NotifierService.record_event')
    @patch('tracker.service.TrackerService.record_event')
    @patch('activitylogs.service.ActivityLogService.record_event')
    def test_tensorboard_failed(self,
                                activitylogs_record,
                                tracker_record,
                                notifier_record,
                                executor_record):
        auditor.record(event_type=tensorboard_events.TENSORBOARD_FAILED,
                       instance=self.tensorboard,
                       target='project')

        assert tracker_record.call_count == 1
        assert activitylogs_record.call_count == 0
        assert notifier_record.call_count == 1
        assert executor_record.call_count == 1

    @patch('executor.executor_service.ExecutorService.record_event')
    @patch('notifier.service.NotifierService.record_event')
    @patch('tracker.service.TrackerService.record_event')
    @patch('activitylogs.service.ActivityLogService.record_event')
    def test_tensorboard_succeeded(self,
                                   activitylogs_record,
                                   tracker_record,
                                   notifier_record,
                                   executor_record):
        auditor.record(event_type=tensorboard_events.TENSORBOARD_SUCCEEDED,
                       instance=self.tensorboard,
                       target='project')

        assert tracker_record.call_count == 1
        assert activitylogs_record.call_count == 0
        assert notifier_record.call_count == 1
        assert executor_record.call_count == 1

    @patch('executor.executor_service.ExecutorService.record_event')
    @patch('notifier.service.NotifierService.record_event')
    @patch('tracker.service.TrackerService.record_event')
    @patch('activitylogs.service.ActivityLogService.record_event')
    def test_tensorboard_statuses_viewed(self,
                                         activitylogs_record,
                                         tracker_record,
                                         notifier_record,
                                         executor_record):
        auditor.record(event_type=tensorboard_events.TENSORBOARD_STATUSES_VIEWED,
                       instance=self.tensorboard,
                       target='project',
                       actor_name='foo',
                       actor_id=1)

        assert tracker_record.call_count == 1
        assert activitylogs_record.call_count == 1
        assert notifier_record.call_count == 0
        assert executor_record.call_count == 0

    @patch('executor.executor_service.ExecutorService.record_event')
    @patch('notifier.service.NotifierService.record_event')
    @patch('tracker.service.TrackerService.record_event')
    @patch('activitylogs.service.ActivityLogService.record_event')
    def test_tensorboard_updated(self,
                                 activitylogs_record,
                                 tracker_record,
                                 notifier_record,
                                 executor_record):
        auditor.record(event_type=tensorboard_events.TENSORBOARD_UPDATED,
                       instance=self.tensorboard,
                       actor_name='foo',
                       actor_id=1)

        assert tracker_record.call_count == 1
        assert activitylogs_record.call_count == 1
        assert notifier_record.call_count == 0
        assert executor_record.call_count == 0

    @patch('executor.executor_service.ExecutorService.record_event')
    @patch('notifier.service.NotifierService.record_event')
    @patch('tracker.service.TrackerService.record_event')
    @patch('activitylogs.service.ActivityLogService.record_event')
    def test_tensorboard_deleted(self,
                                 activitylogs_record,
                                 tracker_record,
                                 notifier_record,
                                 executor_record):
        auditor.record(event_type=tensorboard_events.TENSORBOARD_DELETED,
                       instance=self.tensorboard)

        assert tracker_record.call_count == 1
        assert activitylogs_record.call_count == 0
        assert notifier_record.call_count == 0
        assert executor_record.call_count == 0

    @patch('executor.executor_service.ExecutorService.record_event')
    @patch('notifier.service.NotifierService.record_event')
    @patch('tracker.service.TrackerService.record_event')
    @patch('activitylogs.service.ActivityLogService.record_event')
    def test_tensorboard_triggered_deleted(self,
                                           activitylogs_record,
                                           tracker_record,
                                           notifier_record,
                                           executor_record):
        auditor.record(event_type=tensorboard_events.TENSORBOARD_DELETED_TRIGGERED,
                       instance=self.tensorboard,
                       actor_name='foo',
                       target='project',
                       actor_id=1)

        assert tracker_record.call_count == 1
        assert activitylogs_record.call_count == 1
        assert notifier_record.call_count == 0
        assert executor_record.call_count == 0

    @patch('executor.executor_service.ExecutorService.record_event')
    @patch('notifier.service.NotifierService.record_event')
    @patch('tracker.service.TrackerService.record_event')
    @patch('activitylogs.service.ActivityLogService.record_event')
    def test_tensorboard_job_archived(self,
                                      activitylogs_record,
                                      tracker_record,
                                      notifier_record,
                                      executor_record):
        auditor.record(event_type=tensorboard_events.TENSORBOARD_ARCHIVED,
                       instance=self.tensorboard,
                       actor_name='foo',
                       actor_id=1)

        assert tracker_record.call_count == 1
        assert activitylogs_record.call_count == 1
        assert notifier_record.call_count == 0
        assert executor_record.call_count == 0

    @patch('executor.executor_service.ExecutorService.record_event')
    @patch('notifier.service.NotifierService.record_event')
    @patch('tracker.service.TrackerService.record_event')
    @patch('activitylogs.service.ActivityLogService.record_event')
    def test_tensorboard_job_restored(self,
                                      activitylogs_record,
                                      tracker_record,
                                      notifier_record,
                                      executor_record):
        auditor.record(event_type=tensorboard_events.TENSORBOARD_RESTORED,
                       instance=self.tensorboard,
                       actor_name='foo',
                       actor_id=1)

        assert tracker_record.call_count == 1
        assert activitylogs_record.call_count == 1
        assert notifier_record.call_count == 0
        assert executor_record.call_count == 0


del AuditorBaseTest
